#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

'''
Store API for generic storage of entities
  - orionStore: saves batches to Orion environment
  - sqlFileStore: saves batches to SQL File
'''
from typing import Callable, Dict, Any, Iterable, Sequence, List, Optional
from contextlib import contextmanager
from pathlib import Path
from collections import defaultdict
import json
import itertools
from .cb import cbManager
from .auth import authManager
import psycopg2


# A Store is a callable where you can save batches of entities.
Store = Callable[[Iterable[Any]], None]

@contextmanager
def orionStore(cb: cbManager, auth: authManager, *, service:str=None, subservice:str=None, actionType:str='append', options:list=[]):
    '''
    Context manager that creates a store to save entities to the given cbManager
    All parameters are the same as for the cbManager.send_batch function
    '''
    def send_batch(entities: Iterable[Any]):
        cb.send_batch(service=service, subservice=subservice, auth=auth, actionType=actionType, options=options, entities=entities)
    yield send_batch

@contextmanager
def sqlFileStore(path: Path, *, subservice:str, schema:str=":target_schema", namespace:str="", table_names:Optional[Dict[str, str]]=None, chunk_size:int=10000, append:bool=False, replace_id:Optional[Dict[str, Sequence[str]]]=None):
    '''
    Context manager that creates a store to save entities to an SQL File.
    SQL syntax used is postgresql.

    subservice: name of the orion subservice
    schema: name of the schema. Defaults to `:target_schema` (a psql variable)
    namespace: optional prefix for all tables. Table name for each entity type defaults to "namespace_entityType"
    table_names: dict to override default table names.
      - if table_name[entityType] exists and is not empty, it is used as table name for the entity type.
      - if table_name[entityType] does not exist, the default table name (namespace_entityType) is used.
      - if table_name[entityType] exists and is empty, entities with the given type are not saved.
    chunk_size: maximum lines in a single insert statement. Default=10000
    append: append to the file instead of overwriting.
    replace_id: dictionary of `entity type` => `list of replace_id attributes`.
        For any entity of the given type, the entity id will be replaced by
        the list of values of the specified attributes, concatenated with "_".
        This mimics the behaviour of `replaceId` param in historic URBO-DEPLOYER flows.
    '''
    mode = "a+" if append else "w+"
    handler = path.open(mode=mode, encoding="utf-8")
    some_table_names = table_names or {} # make sure it is not None
    replace_id = replace_id or {} # make sure it is not None
    try:
        def send_batch(entities: Iterable[Any]):
            """Send a batch of entities to the database"""
            for chunk in iter_chunk(entities, chunk_size):
                handler.write(sqlfile_batch(schema=schema, namespace=namespace, table_names=some_table_names, subservice=subservice, replace_id=replace_id, entities=chunk))
                handler.write("\n")
        yield send_batch
    finally:
        handler.close()


def iter_chunk(iterable, chunk_size: int):
    '''
    Chunks an iterable into smaller iterables.
    
    chunk_size: size of the smaller chunks
    '''
    # See https://stackoverflow.com/questions/8991506/iterate-an-iterator-by-chunks-of-n-in-python
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, chunk_size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)

def sql_escape(obj: Any) -> str:
    '''Escapes a value to be used in a SQL string'''
    # See https://github.com/psycopg/psycopg2/issues/331
    adaptor = psycopg2.extensions.adapt(obj)
    if hasattr(adaptor, 'encoding'):
        adaptor.encoding = 'utf-8'
    return adaptor.getquoted().decode('utf-8')

def sqlfile_values(subservice: str, entity: Dict[str, Any], fields: Iterable[str], replace_id:Optional[Sequence[str]]=None) -> str:
    '''
    Generates a string suitable for SQL insert, with all values of the entity
    
    subservice: subservice name
    entity: ngsi entity
    fields: list of fields to save in the entity (omitting id, type)
    replace_id: optional list of attributes to replace the entity's id
    '''
    entityid = entity['id']
    if replace_id is not None and len(replace_id) > 0:
        values = []
        for attr in replace_id:
            if attr == 'id':
                values.append(entityid)
            else:
                values.append(str(entity[attr]['value']))
        entityid = "_".join(values)
    sql = [
        sql_escape(entityid),
        sql_escape(entity['type']),
        sql_escape(subservice),
        "NOW()"
    ]
    for field in fields:
        entry = entity.get(field, {'type':'Text', 'value': None})
        value_untyped = entry.get('value', None)
        value: str
        if value_untyped is None:
            value = "NULL"
        elif 'json' in entry['type']:
            value = sql_escape(json.dumps(value_untyped))
            if 'geo' in entry['type']:
                value = f"ST_GeomFromGeoJSON({value})"
        else:
            value = sql_escape(value_untyped)
        sql.append(value)
    return f"({','.join(sql)})"

def sqlfile_insert(subservice: str, table_name: str, fields: Sequence[str], entities: Iterable[Any], replace_id:Optional[Sequence[str]]=None) -> str:
    '''
    Generate SQL INSERT lines from sequence of entities
    
    subservice: subservice name
    table_name: SQL table name to use
    fields: sequence of attribute names
    entities: iterable of entities
    replace_id: optional list of alues to replace the entity's id
    '''
    return "\n".join((
        f"INSERT INTO {table_name} (entityid,entitytype,fiwareservicepath,recvtime,{','.join(fields)}) VALUES",
        ",\n".join(
            sqlfile_values(subservice, entity, fields, replace_id)
            for entity in entities
        ) + ";"
    ))

def sql_table_name(schema: str, namespace: str, entity_type: str, table_names: Dict[str, str]) -> str:
    '''
    Generates table_name from namespace and entity_type
    
    namespace: namespace to use to prefix entityType
    entity_type: type of entity
    table_names: overrides default names for any entityType
    '''
    default_name = ((namespace + "_") if namespace != "" else "") + entity_type.lower()
    mapped_name  = table_names.get(entity_type, default_name)
    if not mapped_name: # Tables mapped to empty string or None are not saved to SQL
        return ""
    # Tables mapped to some name are prefixed with schema name
    return f"{schema}.{mapped_name}"

def sqlfile_batch(schema: str, namespace: str, table_names: Dict[str, str], subservice: str, replace_id: Dict[str, Sequence[str]], entities: Iterable[Any]) -> str:
    '''
    Generate a single SQL insert batch statement

    schema: schema name
    namespace: namespace for table names
    table_names: overrides entytyType to table name defaults
    subservice: subservice name
    replace_id: map of entity type to list of replace_id attributes
    entities: Iterable of entities
    '''
    # Group entities by type
    entities_by_type = defaultdict(list)
    for entity in entities:
        entities_by_type[entity['type']].append(entity)
    # For each type, get the fieldset
    fields_by_type = defaultdict(set)
    for entity_type, typed_entities in entities_by_type.items():
        for entity in typed_entities:
            for field in entity:
                if field not in ('id', 'type'):
                    fields_by_type[entity_type].add(field)
    # Generate all rows for each type
    sql = []
    for entity_type, typed_entities in sorted(entities_by_type.items()):
        table_name = sql_table_name(schema, namespace, entity_type, table_names)
        if table_name:
            table_cols = sorted(fields_by_type[entity_type])
            sql.append(sqlfile_insert(subservice, table_name, table_cols, typed_entities, replace_id.get(entity_type, None)))
    # and return SQL insert code
    return "\n".join(sql)



# -*- coding: utf-8 -*-
# Copyright 2023 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.
#

import unicodedata
import re

from typing import Mapping, Optional

_whitespace_re = re.compile(r"\s+")

class normalizer:
    """
    Normalizer is a class that will normalize unicode strings to
    valid NGSI entity IDs. Normalization rules are at:

    https://github.com/telefonicaid/fiware-orion/blob/master/doc/manuals/orion-api.md#general-syntax-restrictions

    Normalizers have a __call__ function that takes an input string and:

    - Turn accented characters (á, é, í, ó, u) into unaccented variants.
    - Remove any other unicode character not available in ascii
    - Remove ascii control codes
    - Replace forbidden characters '&', '?', '/', '#' '<', '>', '"', ''', '=', ';', '(', ')'
      with the replacement character (default "-", can be changed in the constructor)
    - Merges consecutive whitespace and replaces it with the replacement character

    You can also set a different replacement character for a specific forbidden
    character, by adding the translation to the `override` optional argument of the
    constructor.

    E.g. if you want to replace " " with "+", you can call:

    ```
    norm = normalizer(override={" ": "+"})
    norm("text (with spaces)")
    ```

    And you will get `"text+-with+spaces-"`.

    You can also remove a forbidden character altogether, by setting its value to
    `None` in the `override` argument. E.g if you want to remove parenthesis,
    you can call:

    ```
    norm = normalizer(override={"(": None, ")": None})
    norm("text (with parenthesis)")
    ```

    If you want to remove ALL special characters (except whitespace):

    ```
    norm = normalizer(replacement="", override={ " ": "-" })
    norm("text (with & special > characters)")
    ```

    And you will get `"text-with-special-characters"`

    The function does not trim the string size to 256 characters, because
    you might want the full normalized original string to store it somewhere
    else before truncating.
    """

    def __init__(self, replacement: str = "-", override: Optional[Mapping[str, str]] = None):
        """Set the default replacement string and custom override mapping"""
        if override is None:
            override = {}
        forbidden_chars = {
            "&": replacement,
            "?": replacement,
            "/": replacement,
            "#": replacement,
            "<": replacement,
            ">": replacement,
            '"': replacement,
            "'": replacement,
            "=": replacement,
            ";": replacement,
            "(": replacement,
            ")": replacement
        }
        source = []
        target = []
        remove = []
        for key, val in forbidden_chars.items():
            custom = override.get(key, val)
            if custom is None or custom == "":
                remove.append(key)
            else:
                if len(custom) > 1:
                    raise ValueError(f"wrong override '{custom}' for char '{key}': must be a single character")
                source.append(key)
                target.append(custom)
        self.space_replacement = override.get(" ", replacement) or ""
        self.table = str.maketrans(
            "".join(source), "".join(target), "".join(remove))

    def __call__(self, text: str) -> str:
        """Normalize text to NGSI entity ID"""
        global _whitespace_re
        ascii = unicodedata.normalize('NFD', text).encode('utf-8').decode('ascii', errors='ignore')
        without_control_chars = "".join(ch for ch in ascii if unicodedata.category(ch)[0] != "C")
        without_specials = without_control_chars.translate(self.table).strip()
        return _whitespace_re.sub(self.space_replacement, without_specials)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

'''
ContextBroker routines for Python:
  - cbManager.send_batch
  - cbManager.get_entities_page
  - cbManager.get_entities
  - cbManager.delete_entities
'''
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from typing import Iterable, List, Any, Optional

import logging
import time
import json

from . import authManager, exceptions

# control urllib3 post and get verify in false
import urllib3, urllib3.exceptions
urllib3.disable_warnings(category=urllib3.exceptions.InsecureRequestWarning)

logger = logging.getLogger(__name__)

class cbManager:
    """ContextBroker Manager

    endpoint: define service endpoint cb (example: https://<service>:<port>)
    timeout: timeout in seconds (default: 10)
    post_retry_connect: number of retries (defaul: 3)
    post_retry_backoff_factor: retry factor delay -> {backoff factor} * (2 ** ({number of total retries} - 1)) (defaul: 20). (ref. urllib3 docs) (default: False)
    sleep_send_batch: sleep X seconds afters send update batch. (default: 0).
    cb_flowcontrol: Opción del Context Broker, que permite un mejor rendimiento en caso de envío masivo de datos (batch updates). Este mecanismo, requiere arrancar el Context Broker con un flag concreto y en las peticiones de envío de datos, añadir esa opción. Referencia en Fiware Orion Docs (default: False)
    block_size: maximum size per batch, in bytes. Default is 800kb and it is not recommended to change.
    batch_size: maximum size per batch, in entities. Default is 0 (no limitiation, other than block_size).
    """

    endpoint: str
    timeout: int = 10
    post_retry_connect: int = 3
    post_retry_backoff_factor: int = 20
    sleep_send_batch: int = 0
    cb_flowcontrol: bool = False
    # Block size is string chars (= bytes, with ASCII codification). This is calculated
    # for 800k (Orion max request size is 1MB,
    # see https://fiware-orion.readthedocs.io/en/master/user/known_limitations.html)
    # and it is not recommended to change it
    block_size = 800000
    # Batch size is entities. Ignored when batch_size = 0.
    # If > 0 , limits the number of entities per batch.
    # We might want to limit the amount of entities per batch for
    # reasons other than block size (e.g to avoid triggering too
    # many subscriptions per batch and stressing cygnus queues)
    batch_size = 0

    def __init__(self,*, endpoint: Optional[str] = None, timeout: int = 10, post_retry_connect: int = 3, post_retry_backoff_factor: int = 20, sleep_send_batch: int = 0, cb_flowcontrol: bool = False, block_size: int = 800000, batch_size: int = 0) -> None:

        if endpoint is None:
            raise ValueError(f'You must define <<endpoint>> in cbManager')

        self.endpoint = endpoint
        self.timeout = timeout
        self.post_retry_connect = post_retry_connect
        self.post_retry_backoff_factor = post_retry_backoff_factor
        self.sleep_send_batch = sleep_send_batch
        self.cb_flowcontrol = cb_flowcontrol
        self.batch_size = batch_size

        # check block_size limit.
        if (int(block_size) > int(800000)):
            raise ValueError('Block size limit reached! <<block_size>> value cannot be greater than 800000')

        self.block_size = block_size


    def delete_entities(self, *, service: str = None, subservice: str = None, auth: authManager = None, limit: int = 100, type: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None, options_get: list = [], options_send: list = []):
        """Delete data from context broker

        :param service: Define service from which entities are deleted, defaults to None
        :param subservice: Define subservice from which entities are deleted, defaults to None
        :param auth: Define authManager, defaults to None
        :param limit: Limits the number of entities to be retrieved, deleted to None
        :param type: Deleted entities whose type matches one of the elements in the list, defaults to None
        :param q: Retrieve entities filtering by attribute value, defaults to None
        :param mq: Retrieve entities filtering by metadata, defaults to None
        :param georel: Georel is intended to specify a spatial relationship between matching entities and a reference shape (geometry), defaults to None
        :param geometry: Allows to define the reference shape to be used when resolving the query. (point | polygon | line | box), defaults to None
        :param coords: Must be a string containing a semicolon-separated list of pairs of geographical coordinates in accordance with the geometry specified, defaults to None
        :param id: Delete entities filtering by Identity, defaults to None
        :param options_get: Options used in Context Broker to find entities, defaults to None
        :param options_send: Options used in Context Broker to delete entities, defaults to None
        :raises ValueError: is thrown when some required argument is missing
        :raises FetchError: is thrown when the response from the cb indicates an error
        """
        data = self.get_entities(service=service, subservice=subservice, auth=auth, limit = limit, type = type, q = q, mq = mq, georel = georel, geometry = geometry, coords = coords, id = id, options=options_get)

        entities = []
        for i, item in enumerate(data):
            entity = {
                'id': f'{item["id"]}',
                'type': f'{item["type"]}',
            }
            entities.append(entity)

        self.send_batch(service=service, subservice=subservice, auth=auth, entities=entities, actionType='delete', options=options_send)

    def get_entities(self, *, service: str = None, subservice: str = None, auth: authManager = None, limit: int = 100, type: str = None, orderBy: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None, options: list = []):
        """Retrieve data from context broker

        :param service: Define service from which entities are retrieved, defaults to None
        :param subservice: Define subservice from which entities are retrieved, defaults to None
        :param auth: Define authManager, defaults to None
        :param limit: Limits the number of entities to be retrieved, defaults to None
        :param type: Retrieve entities whose type matches one of the elements in the list, defaults to None
        :param orderBy: Criteria for ordering results, defaults to None
        :param q: Retrieve entities filtering by attribute value, defaults to None
        :param mq: Retrieve entities filtering by metadata, defaults to None
        :param georel: Georel is intended to specify a spatial relationship between matching entities and a reference shape (geometry), defaults to None
        :param geometry: Allows to define the reference shape to be used when resolving the query. (point | polygon | line | box), defaults to None
        :param coords: Must be a string containing a semicolon-separated list of pairs of geographical coordinates in accordance with the geometry specified, defaults to None
        :param id: Retrieve entities filtering by Identity, defaults to None
        :param options: Options used to retrive entities, defaults to None
        :raises ValueError: is thrown when some required argument is missing
        :raises FetchError: is thrown when the response from the cb indicates an error
        :return: json data
        """

        result = []
        pg = 1

        data = ['go!']
        while (data != []) :
            offset = (pg-1)*limit
            data = self.get_entities_page(service=service, subservice=subservice, auth=auth, offset = offset, limit = limit, type = type, orderBy = orderBy, q = q, mq = mq, georel = georel, geometry = geometry, coords = coords, id = id, options = options)
            pg += 1
            result += data
        return result

    def get_entities_page(self, *, service: Optional[str] = None, subservice: Optional[str] = None, auth: Optional[authManager] = None, offset: Optional[int] = None, limit: Optional[int] = None, type: Optional[str] = None, orderBy: Optional[str] = None, q: Optional[str] = None, mq: Optional[str] = None, georel: Optional[str] = None, geometry: Optional[str] = None, coords: Optional[str] = None, id: Optional[str] = None, options: list = []):
        """Retrieve data from context broker

        :param service: Define service from which entities are retrieved, defaults to None or auth.service defined value
        :param subservice: Define subservice from which entities are retrieved, defaults to None or auth.subservice defined value
        :param auth: Define authManager, defaults to None
        :param offset: Establishes the offset from where entities are retrieved, defaults to None
        :param limit: Limits the number of entities to be retrieved, defaults to None
        :param type: Retrieve entities whose type matches one of the elements in the list, defaults to None
        :param orderBy: Criteria for ordering results, defaults to None
        :param q: Retrieve entities filtering by attribute value, defaults to None
        :param mq: Retrieve entities filtering by metadata, defaults to None
        :param georel: Georel is intended to specify a spatial relationship between matching entities and a reference shape (geometry), defaults to None
        :param geometry: Allows to define the reference shape to be used when resolving the query. (point | polygon | line | box), defaults to None
        :param coords: Must be a string containing a semicolon-separated list of pairs of geographical coordinates in accordance with the geometry specified, defaults to None
        :param id: Retrieve entities filtering by Identity, defaults to None
        :param options: Options used, defaults to None
        :raises ValueError: is thrown when some required argument is missing
        :raises FetchError: is thrown when the response from the cb indicates an error
        :return: json data
        """

        if (auth is not None and not hasattr(auth,"tokens")):
            auth.tokens = {}

        #check subservice defined
        if subservice is None:
            if auth is not None:
                if (not hasattr(auth,"subservice")):
                    raise ValueError('You must define <<subservice>> in authManager')
                else:
                    subservice = auth.subservice
            else:
                raise ValueError('You must define <<subservice>>')
        # auth.subservice is Optional, it might be None
        if subservice is None:
            raise ValueError('You must define <<subservice>> either here or in authManager')

        #check service defined
        if service is None:
            if auth is not None:
                if (not hasattr(auth,"service")):
                    raise ValueError('You must define <<service>> in authManager')
                else:
                    service = auth.service
            else:
                raise ValueError('You must define <<service>>')

        if (auth is not None and subservice not in auth.tokens.keys()):
            auth.get_auth_token_subservice(subservice = subservice)

        headers = {
            'Fiware-Service': service,
            'Fiware-ServicePath': subservice
        }
        if auth is not None:
            headers['X-Auth-Token'] = auth.tokens[subservice]

        # check if use geographical queries, must specify georel, geometry, coords
        if (georel is not None or geometry is not None or coords is not None):
            if (georel is not None and georel != '') and (geometry is not None and geometry != '') and (coords is not None and coords != ''):
                pass
            else:
                raise ValueError('If use geographical queries, you must define georel, geometry and coords in params')


        params = {"offset": offset, "limit": limit, "type": type, "orderBy": orderBy, "q": q, "mq": mq, "georel": georel, "geometry": geometry, "coords": coords, "id": id}

        req_url = ""
        if (options is not None and len(options) > 0):
            req_url = f"{self.endpoint}/v2/entities?options={','.join(options)}"
        else:
            req_url = f"{self.endpoint}/v2/entities"

        resp = requests.get(req_url, params=params, headers=headers, verify=False, timeout=self.timeout)
        if resp.status_code == 400 or resp.status_code == 401:
            respjson = resp.json()
            logger.error(f'{respjson["name"]}: {respjson["message"]}')
        if resp.status_code < 200 or resp.status_code > 204:
            raise exceptions.FetchError(response=resp, method="GET", url=req_url, params=params, headers=headers)

        return resp.json()


    def send_batch(self, *, service:str = None, subservice: str = None, auth: authManager = None, entities: Iterable[Any], actionType: str = 'append', options: list = []) -> bool:
        """Send batch data to context broker with block control

        :param auth: Define authManager
        :param entities: Entities data
        :param service: Define service to send batch data, defaults to None
        :param subservice: Define subservice to send batch data, defaults to None
        :param actionType: Batch action type, defaults is append
        :param options: Options used, defaults to None
        :raises ValueError: is thrown when some required argument is missing
        :raises Exception: is thrown when the cotext broker response isn't ok operation
        :return: True if the operation is correct
        """
        entitiesToSend = []
        accumulated_block = 0
        for entity in entities:
            entitiesToSend.append(entity)
            accumulated_block += len(json.dumps(entity))

            if accumulated_block > self.block_size or (self.batch_size > 0 and len(entitiesToSend) >= self.batch_size) :
                logger.debug(f'- Sending a batch {actionType} of {len(entitiesToSend)} entities')
                # FIXME: This is actually not needed since `__send_batch` never returns False
                # (it either returns True or raises an exception).
                # But technically, both `__send_batch` and `send_batch` return a bool,
                # so we have to manage this and take some action to make the type checker happy.
                ok = self.__send_batch(auth=auth, service=service, subservice=subservice, entities=entitiesToSend, actionType=actionType, options=options)
                if not ok:
                    return False
                entitiesToSend = []
                accumulated_block = 0

        # Remaining block, if any
        if accumulated_block > 0:
            logger.debug(f'- Sending final batch {actionType} of {len(entitiesToSend)} entities')
            return self.__send_batch(auth=auth, service=service, subservice=subservice, entities=entitiesToSend, actionType=actionType, options=options)

        return True

    def __send_batch(self, *, service:str = None, subservice: str = None, auth: authManager = None, entities: List[Any], actionType: str = 'append', options: list = []) -> bool:
        """Send batch data to context broker

        :param auth: Define authManager
        :param entities: Entities data
        :param service: Define service to send batch data, defaults to None
        :param subservice: Define subservice to send batch data, defaults to None  or auth.subservice defined value
        :param actionType: Batch action type, defaults is append
        :param options: Options used, defaults to None
        :raises ValueError: is thrown when some required argument is missing
        :raises Exception: is thrown when the cotext broker response isn't ok operation
        :return: True if the operation is correct
        """

        if (auth is not None and not hasattr(auth,"tokens")):
            auth.tokens = {}

        #check subservice defined
        if subservice is None:
            if auth is not None:
                if not hasattr(auth,"subservice"):
                    raise ValueError('You must define <<subservice>> in authManager')
                else:
                    subservice = auth.subservice
            else:
                raise ValueError('You must define <<subservice>>')

        if (auth is not None and subservice not in auth.tokens.keys()):
            auth.get_auth_token_subservice(subservice = subservice)

        res = self.__batch_creation(auth=auth, service=service, subservice = subservice, entities=entities, actionType=actionType, options=options)
        if (auth is not None and res.status_code == 401):
            auth.get_auth_token_subservice(subservice = subservice)
            res = self.__batch_creation(auth=auth, service=service, subservice = subservice, entities=entities, actionType=actionType, options=options)

        if res.status_code != 204:
            raise Exception(f'Error in batch {actionType} operation ({res.status_code}): {res.json()}')

        logger.debug(f'- Update batch {actionType} of {len(entities)} entities')

        if (self.sleep_send_batch != 0):
            time.sleep(self.sleep_send_batch)

        return True

    def __batch_creation(self, *, service: str = None, subservice: str = None, auth: authManager = None, entities: List[Any], actionType: str = 'append', options: list = []):
        """Send batch data to Context Broker

        :param entities: Entities data
        :param service: Define service to send batch data, defaults to None or auth.service defined value
        :param subservice: Define subservice to send batch data, defaults to None or auth.subservice defined value
        :param auth: Define authManager, defaults to None
        :param actionType: Batch action type, defaults is append
        :params options: Options used, default to None
        :raises ValueError: is thrown when some required argument is missing
        :return: Http response code
        """

        #check subservice defined
        if subservice is None:
            if auth is not None:
                if not hasattr(auth,"subservice"):
                    raise ValueError('You must define <<subservice>> in authManager')
                else:
                    subservice = auth.subservice
            else:
                raise ValueError('You must define <<subservice>>')
        # auth.subservice is Optional, it could still be None
        if subservice is None:
            raise ValueError('You must define <<subservice>> either here or in authManager')

        #check service defined
        if service is None:
            if auth is not None:
                if not hasattr(auth,"service"):
                    raise ValueError('You must define <<service>> in authManager')
                else:
                    service = auth.service
            else:
                raise ValueError('You must define <<service>>')


        if (not hasattr(self,"endpoint")):
            raise ValueError('You must define <<endpoint>> in cbManager')

        headers = {
            'Fiware-Service': service,
            'Fiware-ServicePath': subservice,
            'Content-Type': 'application/json'
        }
        if (auth is not None):
            headers['X-Auth-Token'] = auth.tokens[subservice]

        body = {
            'actionType': f'{actionType}',
            'entities': entities
        }

        # if cb_flowcontrol, add flowControl flag to options
        if (self.cb_flowcontrol):
            if (options is None):
                options = ['flowControl']
            else:
                # check if flowcontrol is in options.
                if 'flowControl' not in options:
                    options.append('flowControl')

        if (options is not None and len(options) > 0):
            req_url = f"{self.endpoint}/v2/op/update?options={','.join(options)}"
        else:
            req_url = f"{self.endpoint}/v2/op/update"

        http = requests.Session()
        retry_strategy = Retry(
            total=self.post_retry_connect,
            read=self.post_retry_connect,
            backoff_factor=self.post_retry_backoff_factor,
            status_forcelist=(429, 500, 502, 503, 504),
            method_whitelist=('HEAD', 'GET', 'OPTIONS', 'POST')
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        http.mount('http://', adapter)
        http.mount('https://', adapter)
        return http.post(req_url, json=body, headers=headers, verify=False, timeout=self.timeout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

"""
Authorization routines for Python:
  - authManager.get_auth_token_subservice
"""

import requests
import logging
from typing import Dict, Optional, cast

logger = logging.getLogger(__name__)

class authManager:
    """Authentication Manager

    endpoint: define service endpoint auth (example: https://<service>:<port>)
    user: define user to authenticate
    password: define password to authenticate
    service: service to authenticate
    subservice: subservice to authenticate
    tokens: token list stored when athenticate in a subservice
    """
    endpoint: str
    user: str
    password: str
    service: str
    subservice: Optional[str]
    tokens: Dict[str, str]

    def __init__(self, *, endpoint: Optional[str] = None, service: Optional[str] = None, user: Optional[str] = None, password: Optional[str] = None,
                 subservice: Optional[str] = None) -> None:

        messageError = []
        if endpoint is None:
            messageError.append('<<endpoint>>')

        if service is None:
            messageError.append('<<service>>')

        if user is None:
            messageError.append('<<user>>')

        if password is None:
            messageError.append('<<password>>')

        if len(messageError) != 0:
            defineParams = messageError[0]
            if len(messageError) != 1:
                defineParams = " and ".join([", ".join(messageError[:-1]), messageError[-1]])
            raise ValueError(f'You must define {defineParams} in authManager')

        # At this point, all Optional[str] have been validated to be not None.
        # cast them to let type checker knows.
        self.endpoint = cast(str, endpoint)
        self.service = cast(str, service)
        self.user = cast(str, user)
        self.password = cast(str, password)
        self.subservice = subservice
        self.tokens = {}

    def set_token(self, subservice: str, token: str):
        self.subservice = subservice
        self.tokens[subservice] = token

    def get_info(self):
        """ Show auth info

        only for debug uses
        """
        if (hasattr(self, "endpoint")):
            logger.info(f'endpoint: {self.endpoint}')
        if (hasattr(self, "service")):
            logger.info(f'service: {self.service}')
        if (hasattr(self, "subservice")):
            logger.info(f'service: {self.subservice}')
        if (hasattr(self, "user")):
            logger.info(f'user: {self.user}')
        if (hasattr(self, "password")):
            logger.info(f'password: {self.password}')
        if (hasattr(self, "tokens")):
            for key in self.tokens:
                logger.info('tokens[' + key + ']: ' + self.tokens[key])

    def check_mandatory_fields(self):
        """Raise ValueError if some mandatory field is missing"""
        messageError = []
        if (not hasattr(self, "endpoint")):
            messageError.append('<<endpoint>>')

        if (not hasattr(self, "service")):
            messageError.append('<<service>>')

        if (not hasattr(self, "user")):
            messageError.append('<<user>>')

        if (not hasattr(self, "password")):
            messageError.append('<<password>>')

        if len(messageError) != 0:
            defineParams = messageError[0]
            if len(messageError) != 1:
                defineParams = " and ".join([", ".join(messageError[:-1]), messageError[-1]])
            raise ValueError(f'You must define {defineParams} in authManager')

    def _post_auth_request(self, detail: str, scope: dict):
        """Send a POST Auth request to keystone for the given scope

        :param detail: a detail string to include in logging messages
        :param scope: the scope object to send in the request, in keystone v3 format
        :raises Exception: when authorization fails
        :return: requests.Response
        """
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        body = {
            "auth": {
                "scope": scope,
                "identity": {
                    "password": {
                        "user": {
                            "domain": {
                                "name": self.service
                            },
                            "password": self.password,
                            "name": self.user
                        }
                    },
                    "methods": [
                        "password"
                    ]
                }
            }
        }

        logger.debug(f'getting auth token ({detail})...')
        req_url = self.endpoint + '/v3/auth/tokens'
        res = requests.post(req_url, json=body, headers=headers, verify=False)

        if res.status_code != 201:
            raise Exception(f'Failed to get auth token ({detail}) ({res.status_code}): {res.json()}')

        logger.debug(f'Authentication token ({detail}) was created successfully')
        return res

    def get_auth_token_subservice(self, *, subservice: Optional[str] = None):
        """Authenticate in a service/subservice and get a token

        :param subservice: define subservice to be authenticated, defaults to None
        :raises ValueError: is thrown when some required argument is missing
        :raises Exception: is thrown when the response from the auth service indicates an error
        :return: the token returned by the auth service
        """

        self.check_mandatory_fields()
        if (not hasattr(self, "tokens")):
            self.tokens = {}

        if subservice is None:
            if (not hasattr(self, "subservice")):
                raise ValueError('You must define <<subservice>> in authManager')
            else:
                subservice = self.subservice
        # auth.subservice is Optional, it might be None
        if subservice is None:
            raise ValueError('You must define <<subservice>>')

        res = self._post_auth_request(detail=f"subservice {subservice}", scope={
            "project": {
                "domain": {
                    "name": self.service
                },
                "name": subservice
            }
        })

        token = res.headers['X-Subject-Token']
        self.tokens[subservice] = token
        return token

    def get_auth_token_service(self):
        """Authenticate in a service and get a token.
        The service token is NOT cached.

        :raises ValueError: is thrown when some required argument is missing
        :raises Exception: is thrown when the response from the auth service indicates an error
        :return: an object with { 'token': .... 'user_id': ..., 'domain_id': ... }
        """

        self.check_mandatory_fields()
        res = self._post_auth_request(detail=f"service {self.service}", scope={
            "domain": {
                "name": self.service
            }
        })

        token = res.headers['X-Subject-Token']
        tbody = res.json()['token']
        return {
            'token': token,
            'user_id': tbody['user']['id'],
            'domain_id': tbody['user']['domain']['id'],
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

"""
Exceptions handling for Python.
"""

import requests
from typing import Any, Optional

class FetchError(Exception):
    """
    FetchError encapsulates all parameters of an HTTP request and the erroneous response.
    """

    response: requests.Response
    method: str
    url: str
    params: Optional[Any] = None
    headers: Optional[Any] = None
    body: Optional[Any] = None

    def __init__(self, response: requests.Response, method: str, url: str, params: Optional[Any] = None, headers: Optional[Any] = None, body: Optional[Any] = None):
        """Constructor for FetchError class."""
        self.response = response
        self.method = method
        self.url = url
        self.params = params
        self.headers = headers
        self.body = body

    def __str__(self) -> str:
        return f"Failed to {self.method} {self.url} (headers: {self.headers}, params: {self.params}, body: {self.body}): [{self.response.status_code}] {self.response.text}"


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

'''
SQLFile Store tests
'''

import unittest
import textwrap
import os
import tempfile
from pathlib import Path
from tc_etl_lib import sqlFileStore

TestEntities = [
    {
        "id": "id_1",
        "type": "type_A",
        "municipality": {
            "type": "Text",
            "value": "NA"
        },
        "location": {
            "type": "geo:json",
            "value": { "type": "Point", "coordinates": [1, 2] }
        }
    },
    {
        "id": "id_2",
        "type": "type_B",
        "TimeInstant": {
            "type": "DateTime",
            "value": "2022-12-15T18:00:00Z"
        },
        "temperature": {
            "type": "Number",
            "value": 20
        }
    },
    {
        "id": "id_3",
        "type": "type_A",
        "municipality": {
            "type": "Text",
            "value": "NA"
        },
        "location": {
            "type": "geo:json",
            "value": { "type": "Point", "coordinates": [3, 4] }
        }
    },
    {
        "id": "id_4",
        "type": "type_B",
        "TimeInstant": {
            "type": "DateTime",
            "value": "2022-12-15T18:01:00Z"
        },
        "temperature": {
            "type": "Number",
            "value": 21
        }
    },
    {
        "id": "id_5",
        "type": "type_A",
        "municipality": {
            "type": "Text",
            "value": "Alcobendas"
        },
        "location": {
            "type": "geo:json",
            "value": { "type": "Point", "coordinates": [5, 6] }
        }
    },
]


def dedent(text: str) -> str:
    '''Remove leading whitespace'''
    return textwrap.dedent(text.strip("\n"))

class TestSQLFileStore(unittest.TestCase):
    '''Tests for sqlFileStore'''

    def do_test(self, expect: str, append_text: str, entities=TestEntities, **params):
        '''test sqlFileStore with given parameters and expectations'''
        with tempfile.NamedTemporaryFile(mode='w+', encoding='utf-8', delete=False) as tmpFile:
            try:
                tmpFile.close() # we only want the name, not the handle
                if append_text != "":
                    with open(tmpFile.name, "w+", encoding="utf-8") as outfile:
                        outfile.write(dedent(append_text))
                with sqlFileStore(Path(tmpFile.name), **params) as store:
                    store(entities)
                with open(tmpFile.name, "r", encoding="utf-8") as infile:
                    data = infile.read()
                    self.maxDiff = None
                    self.assertEqual(dedent(data).strip(), dedent(expect).strip())
            finally:
                os.unlink(tmpFile.name)

    def test_default_tables(self):
        '''Test SQL File using standard table names'''
        expected = """
        INSERT INTO :target_schema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_1','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA'),
        ('id_3','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [3, 4]}'),'NA'),
        ('id_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [5, 6]}'),'Alcobendas');
        INSERT INTO :target_schema.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20),
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, "", subservice="/testsrv", chunk_size=5)

    def test_schema_tables(self):
        '''Test SQL File using standard table names and custom schema'''
        expected = """
        INSERT INTO schema1.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_1','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA'),
        ('id_3','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [3, 4]}'),'NA'),
        ('id_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [5, 6]}'),'Alcobendas');
        INSERT INTO schema1.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20),
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, "", subservice="/testsrv", schema="schema1", chunk_size=5)

    def test_namespaced_tables(self):
        '''Test SQL File using namespaced names'''
        expected = """
        INSERT INTO :target_schema.ns_type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_1','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA'),
        ('id_3','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [3, 4]}'),'NA'),
        ('id_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [5, 6]}'),'Alcobendas');
        INSERT INTO :target_schema.ns_type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20),
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, "", namespace="ns", subservice="/testsrv")

    def test_rename_entity(self):
        '''Test SQL File overriding some entity tables'''
        expected = """
        INSERT INTO :target_schema.renamed (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_1','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA'),
        ('id_3','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [3, 4]}'),'NA'),
        ('id_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [5, 6]}'),'Alcobendas');
        INSERT INTO :target_schema.ns_type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20),
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, "", subservice="/testsrv", namespace="ns", table_names={'type_A': 'renamed'})

    def test_skip_entity(self):
        '''Test SQL File skipping some entity types'''
        expected = """
        INSERT INTO :target_schema.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20),
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, "", subservice="/testsrv", table_names={'type_A': None})

    def test_small_chunks(self):
        '''Test SQL File splitting batches in smaller chunks'''
        expected = """
        INSERT INTO :target_schema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_1','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA'),
        ('id_3','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [3, 4]}'),'NA');
        INSERT INTO :target_schema.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20);
        INSERT INTO :target_schema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [5, 6]}'),'Alcobendas');
        INSERT INTO :target_schema.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, "", subservice="/testsrv", chunk_size=3)

    def test_append(self):
        '''Test SQL File append to existing data'''
        create = """
        CREATE TABLE IF NOT EXISTS myschema.type_a (entityid text, entitytype text, fiwareservicepath text, recvtime timestamp with time zone not null, location geometry, municipality text);
        CREATE TABLE IF NOT EXISTS myschema.type_b (entityid text, entitytype text, fiwareservicepath text, recvtime timestamp with time zone not null, timeinstant timestamp with time zone not null, temperature doule precision);
        """
        expected = """
        CREATE TABLE IF NOT EXISTS myschema.type_a (entityid text, entitytype text, fiwareservicepath text, recvtime timestamp with time zone not null, location geometry, municipality text);
        CREATE TABLE IF NOT EXISTS myschema.type_b (entityid text, entitytype text, fiwareservicepath text, recvtime timestamp with time zone not null, timeinstant timestamp with time zone not null, temperature doule precision);
        INSERT INTO myschema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_1','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA'),
        ('id_3','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [3, 4]}'),'NA');
        INSERT INTO myschema.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_2','type_B','/testsrv',NOW(),'2022-12-15T18:00:00Z',20);
        INSERT INTO myschema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality) VALUES
        ('id_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [5, 6]}'),'Alcobendas');
        INSERT INTO myschema.type_b (entityid,entitytype,fiwareservicepath,recvtime,TimeInstant,temperature) VALUES
        ('id_4','type_B','/testsrv',NOW(),'2022-12-15T18:01:00Z',21);
        """
        self.do_test(expected, create, subservice="/testsrv", schema="myschema", chunk_size=3, append=True)

    def test_singleton_one_id(self):
        '''Test replace_id with a single attribute'''
        entities  = [
            {
                "id": "id_singleton",
                "type": "type_A",
                "ref": {
                    "type": "Text",
                    "value": "id_A"
                },
                "municipality": {
                    "type": "Text",
                    "value": "NA"
                },
                "location": {
                    "type": "geo:json",
                    "value": { "type": "Point", "coordinates": [1, 2] }
                }
            },
            {
                "id": "id_no_singleton",
                "type": "type_B",
                "ref": {
                    "type": "Text",
                    "value": "id_B"
                },
                "municipality": {
                    "type": "Text",
                    "value": "NA"
                },
                "location": {
                    "type": "geo:json",
                    "value": { "type": "Point", "coordinates": [1, 2] }
                }
            }
        ]
        expected = """
        INSERT INTO myschema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality,ref) VALUES
        ('id_A','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA','id_A');
        INSERT INTO myschema.type_b (entityid,entitytype,fiwareservicepath,recvtime,location,municipality,ref) VALUES
        ('id_no_singleton','type_B','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA','id_B');
        """
        self.do_test(expected, "", entities=entities, replace_id={'type_A':['ref']}, subservice="/testsrv", schema="myschema")

    def test_singleton_two_ids(self):
        '''Test replace_id with two attributes'''
        entities  = [
            {
                "id": "id_singleton",
                "type": "type_A",
                "primary": {
                    "type": "Text",
                    "value": "id_primary"
                },
                "secondary": {
                    "type": "Number",
                    "value": 5
                },
                "municipality": {
                    "type": "Text",
                    "value": "NA"
                },
                "location": {
                    "type": "geo:json",
                    "value": { "type": "Point", "coordinates": [1, 2] }
                }
            }
        ]
        expected = """
        INSERT INTO myschema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality,primary,secondary) VALUES
        ('id_primary_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA','id_primary',5);
        """
        self.do_test(expected, "", entities=entities, replace_id={'type_A':['primary', 'secondary']}, subservice="/testsrv", schema="myschema")

    def test_singleton_original_id(self):
        '''Test replace_id with the original id, plus some attrib'''
        entities  = [
            {
                "id": "id_singleton",
                "type": "type_A",
                "ref": {
                    "type": "Number",
                    "value": 5
                },
                "municipality": {
                    "type": "Text",
                    "value": "NA"
                },
                "location": {
                    "type": "geo:json",
                    "value": { "type": "Point", "coordinates": [1, 2] }
                }
            }
        ]
        expected = """
        INSERT INTO myschema.type_a (entityid,entitytype,fiwareservicepath,recvtime,location,municipality,ref) VALUES
        ('id_singleton_5','type_A','/testsrv',NOW(),ST_GeomFromGeoJSON('{"type": "Point", "coordinates": [1, 2]}'),'NA',5);
        """
        self.do_test(expected, "", entities=entities, replace_id={'type_A':['id', 'ref']}, subservice="/testsrv", schema="myschema")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

'''
IoT Agent routines for Python:
  - iotaManager.send_http
  - iotaManager.send_batch_http
'''

from . import exceptions
import pandas as pd
import requests
import tc_etl_lib as tc
import time
from typing import Iterable, Optional, Union

class SendBatchError(Exception):
    "SendBatchError is a class that can handle exceptions."
    def __init__(self, message, original_exception=None, index=None):
        super().__init__(message)
        self.original_exception = original_exception
        self.index = index

class iotaManager:
    """IoT Agent Manager.

    endpoint: define service endpoint iota (example: https://<service>:<port>).
    sensor_id: sensor ID.
    api_key: API key of the corresponding sensor.
    sleep_send_batch: time sleep in seconds (default: 0).
    """

    endpoint: str
    sensor_id: str
    api_key: str
    sleep_send_batch: float

    def __init__(self, endpoint: str, sensor_id: str, api_key: str, sleep_send_batch: float = 0):
        self.endpoint = endpoint
        self.sensor_id = sensor_id
        self.api_key = api_key
        self.sleep_send_batch = sleep_send_batch

    def send_http(self,
                data: dict) -> Union[None, bool]:

        if not isinstance(data, dict):
                raise TypeError("The 'data' parameter should be a dictionary with key-value pairs.")

        params = {
            'i': self.sensor_id,
            'k': self.api_key
        }
        headers = {
            "Content-Type": "application/json"
        }

        try:
            resp = requests.post(url=self.endpoint, json=data, params=params, headers=headers)
            if resp.status_code == 200:
                return True
            else:
                raise exceptions.FetchError(
                    response=resp,
                    method="POST",
                    url=self.endpoint,
                    params=params,
                    headers=headers)
        except requests.exceptions.ConnectionError as e:
            raise e

    def send_batch_http(self, data: Iterable) -> Union[None, bool]:

            if isinstance(data, pd.DataFrame):
                # Convert each row of the DataFrame to a dictionary.
                for i, row in data.iterrows():
                    try:
                        self.send_http(row.to_dict())
                        time.sleep(self.sleep_send_batch)
                    except Exception as e:
                        raise SendBatchError(f"send_batch_http error. Row that caused the error: {i}\nError detail: {str(e)}", original_exception=e, index=i) from e
            else:
                for i, dictionary in enumerate(data):
                    try:
                        self.send_http(dictionary)
                        time.sleep(self.sleep_send_batch)
                    except Exception as e:
                        raise SendBatchError(f"send_batch_http error. Index where the error occurred: {i}\nError detail: {str(e)}", original_exception=e, index=i) from e
            return True


# -*- coding: utf-8 -*-
#
# Copyright 2023 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

'''
Normalizer tests
'''

import unittest
from tc_etl_lib import normalizer


class TestNormalizer(unittest.TestCase):
    '''Tests for sqlFileStore'''

    def do_test(self, replacement="-", override=None, input="", expected=""):
        '''test Normalizer with the given options dict'''
        norm = normalizer(replacement=replacement, override=override)
        result = norm(input)
        self.assertEqual(result, expected)

    def test_default_behaviour(self):
        self.do_test(
            replacement="-",
            override=None,
            input="text (with & specials) > áéíóúñ",
            expected="text--with---specials----aeioun"
        )

    def test_different_replacement(self):
        self.do_test(
            replacement="_",
            override=None,
            input="text (with & specials) > áéíóúñ",
            expected="text__with___specials____aeioun"
        )

    def test_space_override(self):
        self.do_test(
            replacement="-",
            override={" ": "+"},
            input="text (with & specials) > áéíóúñ",
            expected="text+-with+-+specials-+-+aeioun"
        )

    def test_forbidden_override(self):
        self.do_test(
            replacement="-",
            override={">": "+"},
            input="text (with & specials) > áéíóúñ",
            expected="text--with---specials--+-aeioun"
        )

    def test_forbidden_remove_some(self):
        self.do_test(
            replacement="-",
            override= {
                "(": None,
                ")": None,
            },
            input="text (with & specials) > áéíóúñ",
            expected="text-with---specials---aeioun"
        )

    def test_space_remove(self):
        self.do_test(
            replacement="-",
            override= {" ": None },
            input="text (with & specials) > áéíóúñ",
            expected="text-with-specials--aeioun"
        )

    def test_forbidden_remove_all(self):
        self.do_test(
            replacement="",
            override={ " ": "-" },
            input="text (with & specials) > áéíóúñ",
            expected="text-with-specials-aeioun"
        )

    def test_remove_all(self):
        self.do_test(
            replacement="",
            override=None,
            input="text (with & specials) > áéíóúñ",
            expected="textwithspecialsaeioun"
        )



# -*- coding: utf-8 -*-
#
# Copyright 2023 Telefónica Soluciones de Informática y Comunicaciones de España, S.A.U.
#
# This file is part of tc_etl_lib
#
# tc_etl_lib is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# tc_etl_lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with IoT orchestrator. If not, see http://www.gnu.org/licenses/.

'''
IoT Agent tests.
'''

from . import exceptions
from tc_etl_lib.iota import SendBatchError, iotaManager
import pandas as pd
import pytest
import requests
import unittest
from unittest.mock import patch, Mock, MagicMock


class TestIotaManager(unittest.TestCase):
    def test_send_http_success(self):
        """A success message should be displayed when
        HTTP request is executed successfully."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with patch('requests.post') as mock_post:
            fake_response = Mock()
            # Simulates a successful code status.
            fake_response.status_code = 200
            mock_post.return_value = fake_response
            resp = iot.send_http(data={"key": "value"})
            assert resp == True

    def test_send_http_connection_error(self):
        """Should raise an exception when there is a server connection error."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with patch('requests.post') as mock_post:
            mock_post.side_effect = requests.exceptions.ConnectionError()
            with pytest.raises(requests.exceptions.ConnectionError):
                iot.send_http(data={"key": "value"})

    def test_send_http_invalid_data_type(self):
        """Should raise an exception when the data type is incorrect."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with pytest.raises(TypeError) as exc_info:
            iot.send_http(data=["data"])
        exception_message = str(exc_info.value)
        assert "The 'data' parameter should be a dictionary with key-value pairs." in str(
            exception_message)

    def test_send_http_set_not_unique(self):
        """Should raise an exception if the data is an array of dictionaries."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with pytest.raises(TypeError) as exc_info:
            iot.send_http(data=[
                            {"key_1": "value_1"},
                            {"key_2": "value_2"}])
        exception_message = str(exc_info.value)
        assert "The 'data' parameter should be a dictionary with key-value pairs." in str(
            exception_message)

    def test_send_http_unauthorized(self):
        """Should raise an exception when the request is unauthorized."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with patch('requests.post') as mock_post:
            mock_post.return_value.status_code = 401
            with pytest.raises(exceptions.FetchError) as exc_info:
                iot.send_http(data={"key": "value"})
            exception_raised = str(exc_info.value)
            assert "401" in exception_raised
            assert "Failed to POST http://fakeurl.com" in exception_raised

    def test_send_http_not_found(self):
        """Should raise an exception when the request is not found."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with patch('requests.post') as mock_post:
            mock_post.return_value.status_code = 404
            with pytest.raises(exceptions.FetchError) as exc_info:
                iot.send_http(data={"key": "value"})

            exception_raised = str(exc_info.value)
            assert "404" in exception_raised
            assert "Failed to POST http://fakeurl.com" in exception_raised

    def test_send_http_server_error(self):
        """Should raise an exception if there is a server error."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key')
        with patch('requests.post') as mock_post:
            mock_post.return_value.status_code = 500
            with pytest.raises(exceptions.FetchError) as exc_info:
                iot.send_http(data={"key": "value"})

            exception_raised = str(exc_info.value)
            assert "500" in exception_raised
            assert "Failed to POST http://fakeurl.com" in exception_raised

    def test_send_batch_http_dict_success(self):
        """send_http should be called twice when we send an array with 2 dictionaries."""
        with patch('tc_etl_lib.iota.iotaManager.send_http') as mock_send_http:
            iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key', sleep_send_batch=0.25)
            iot.send_batch_http(data=[{'key1': 'value1'}, {'key2': 'value2'}])
        self.assertEqual(mock_send_http.call_count, 2)

    def test_send_batch_http_dict_value_error(self):
        """Should raise TypeError and then raise SendBatchError with the index that failed."""
        iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key', sleep_send_batch=0.25)
        with patch('requests.post') as mock_post:
            mock_post.return_value.status_code = 200
            with self.assertRaises(SendBatchError) as context:
                iot.send_batch_http(data=[{"key_1": "value_1"}, 2])

        expected_message = "send_batch_http error. Index where the error occurred: 1\nError detail: The 'data' parameter should be a dictionary with key-value pairs."
        self.assertEqual(str(context.exception), expected_message)
        # Verify that the POST request was called for the first dictionary.
        mock_post.assert_called_once()

    def test_send_batch_http_connection_error(self):
            """Should raise a ConnectionError exception and then raise SedBatchError with the index that failed."""
            iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key', sleep_send_batch=0.25)
            with patch('requests.post') as mock_post:
                mock_success = MagicMock()
                mock_success.status_code = 200

                mock_connection_error = MagicMock(side_effect=ConnectionError("Simulated connection error"))
                # Simulates a connection error when tries to send the second dictionary.
                mock_post.side_effect = [mock_success, mock_connection_error]

                with self.assertRaises(SendBatchError) as exc_info:
                    iot.send_batch_http(data=[{"key_1": "value_1"}, {"key_2": "value_2"}])

            exception_raised = str(exc_info.exception)

            assert "send_batch_http error. Index where the error occurred: 1\nError detail:" in exception_raised
            assert "Failed to POST" in exception_raised
            # Verify that it tried to do a POST request for each dictionary.
            assert mock_post.call_count == 2

    def test_send_batch_http_data_frame_success(self):
        """send_http should be called 3 times when we send a DataFrame with 3 rows."""
        with patch('tc_etl_lib.iota.iotaManager.send_http') as mock_send_http:
            iot = iotaManager(endpoint='http://fakeurl.com', sensor_id='fake_sensor_id', api_key='fake_api_key', sleep_send_batch=0.25)
            data = pd.DataFrame({
            'column1': [1, 2, 3],
            'column2': ['a', 'b', 'c']
            })
            iot.send_batch_http(data=data)
        self.assertEqual(mock_send_http.call_count, 3)